#include <err.h>
#include <fcntl.h>
#include <gelf.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <vis.h>

int
main(int argc, char **argv)
{
        int fd, elferr;
        Elf *e;
        char *name, *p;
        Elf_Scn *scn;
        Elf_Data *data;
        GElf_Shdr shdr;
        size_t n, shstrndx;

        if (argc != 2)
                errx(EX_USAGE, "usage: %s file-name", getprogname());

        if (elf_version(EV_CURRENT) == EV_NONE)
                errx(EX_SOFTWARE, "ELF library initialization failed: %s",
                    elf_errmsg(-1));

        if ((fd = open(argv[1], O_RDONLY, 0)) < 0)
                err(EX_NOINPUT, "open \%s\" failed", argv[1]);

        if ((e = elf_begin(fd, ELF_C_READ, NULL)) == NULL)
                errx(EX_SOFTWARE, "elf_begin() failed: %s.",
                    elf_errmsg(-1));

        if (elf_kind(e) != ELF_K_ELF)
                errx(EX_DATAERR, "%s is not an ELF object.", argv[1]);

        if (elf_getshstrndx(e, &shstrndx) == 0)
                errx(EX_SOFTWARE, "getshstrndx() failed: %s.",
                    elf_errmsg(-1));

        scn = NULL;
        while ((scn = elf_nextscn(e, scn)) != NULL) {
                if (gelf_getshdr(scn, &shdr) != &shdr)
                        errx(EX_SOFTWARE, "getshdr() failed: %s.",
                            elf_errmsg(-1));

                if ((name = elf_strptr(e, shstrndx, shdr.sh_name)) == NULL)
                        errx(EX_SOFTWARE, "elf_strptr() failed: %s.",
                            elf_errmsg(-1));

		if (strcmp(name, ".bss") == 0)
			(void) printf(".bss section:\n");
		else
			continue;

		data = NULL; n = 0;
		while (n < shdr.sh_size && (data = elf_getdata(scn, data)) !=
		    NULL) {
			printf("d_buf\t\t=\t%p\n", data->d_buf);
			printf("d_off\t\t=\t%ju\n", data->d_off);
			printf("d_align\t\t=\t%ju\n", data->d_align);
			printf("d_size\t\t=\t%ju\n", data->d_size);
			printf("d_type\t\t=\t%d\n", data->d_type);
			printf("d_version\t=\t%d\n", data->d_version);
			if (data->d_buf == NULL)
				break;
			p = (char *) data->d_buf;
			while (p < (char *) data->d_buf + data->d_size) {
				printf("%c", *p);
				n++; p++;
				(void) putchar((n % 16) ? ' ' : '\n');
			}
		}
		elferr = elf_errno();
		if (elferr != 0)
			errx(EX_SOFTWARE, "elf_getdata failed: %s",
			    elf_errmsg(elferr));
		
		(void) putchar('\n');
        }

        (void) elf_end(e);
        (void) close(fd);
        exit(EX_OK);
}
